#!/usr/bin/env python3

import argparse
import glob
import sys

import numpy as np

from ucca import layer0, layer1
from ucca.ioutil import file2passage, read_files_and_dirs
from ucca.layer1 import NodeTags
from ucca.textutil import break2sentences

desc = """Prints statistics on UCCA passages
"""


def main():
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="files to process")
    argparser.add_argument("-o", "--outfile", help="output file for data")
    argparser.add_argument("-s", "--summary", help="show only summary", action="store_true")
    args = argparser.parse_args()

    print("id,passages,paragraphs,sentences,nodes,terminals,non-terminals,implicit,linkage,discont,"
          "edges,primary,remote,linkage,parents,children,mult-parents")
    data = []
    for pattern in args.filenames:
        for passage in read_files_and_dirs(glob.glob(pattern)):
            terminals = passage.layer(layer0.LAYER_ID).all
            non_terminals = [n for n in passage.layer(layer1.LAYER_ID).all if n.ID != "1.1"]
            non_linkage = [n for n in non_terminals if n.tag != NodeTags.Linkage]
            linkage_nodes = passage.layer(layer1.LAYER_ID).top_linkages
            edges = {e for n in non_terminals for e in n}
            remote = [e for e in edges if e.attrib.get("remote")]
            linkage_edges = [e for n in linkage_nodes for e in n]
            fields = (int(passage.ID),
                      1,
                      len({t.paragraph for t in terminals}),
                      len(break2sentences(passage)),
                      len(terminals) + len(non_terminals),
                      len(terminals),
                      len(non_terminals),
                      len([n for n in non_linkage if n.attrib.get("implicit")]),
                      len(linkage_nodes),
                      len([n for n in non_linkage if n.tag == NodeTags.Foundational and n.discontiguous]),
                      len(edges),
                      len(edges) - len(remote) - len(linkage_edges),
                      len(remote),
                      len(linkage_edges),
                      sum(len([p for p in n.parents if p.ID != "1.1"]) for n in non_linkage),
                      sum(len(n.children) for n in non_linkage),
                      len([n for n in non_linkage if len([p for p in n.parents if p.ID != "1.1"]) > 1]),
                      )
            if not args.summary:
                print(",".join("%d" % f for f in fields))
            data.append(fields)
    data = np.array(data, dtype=int)
    if args.outfile:
        np.savetxt(args.outfile, data[data[:, 0].argsort()], fmt="%i", delimiter="\t")
    if args.summary:
        print(",".join("%d" % f for f in data.sum(axis=0)))

    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import argparse
import os
import sys

from ucca.ioutil import file2passage, passage2file

desc = """Parses an XML in UCCA standard format, and writes them in binary Pickle format.
"""


def main():
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument('filenames', nargs='+', help="XML file names to convert")
    argparser.add_argument('-o', '--outdir', default='.', help="output directory")
    args = argparser.parse_args()

    for filename in args.filenames:
        sys.stderr.write("Reading passage '%s'...\n" % filename)
        passage = file2passage(filename)
        basename = os.path.splitext(os.path.basename(filename))[0]
        outfile = args.outdir + os.path.sep + basename + ".pickle"
        sys.stderr.write("Writing file '%s'...\n" % outfile)
        passage2file(passage, outfile, binary=True)

    sys.exit(0)


if __name__ == '__main__':
    main()



"""Utility functions for UCCA package."""
import os
from itertools import groupby
from operator import attrgetter

import numpy as np
from tqdm import tqdm

from ucca import layer0, layer1

MODEL_ENV_VAR = "SPACY_MODEL"
DEFAULT_MODEL = {"en": "en_core_web_md", "fr": "fr_core_news_md", "de": "de_core_news_sm"}


def get_nlp(lang="en"):
    instance = nlp.get(lang)
    if instance is None:
        import spacy
        model_name = os.environ.get("_".join((MODEL_ENV_VAR, lang.upper()))) or os.environ.get(MODEL_ENV_VAR) or \
            DEFAULT_MODEL.get(lang, "xx")
        try:
            nlp[lang] = instance = spacy.load(model_name)
        except OSError:
            spacy.cli.download(None, model_name)
            try:
                nlp[lang] = instance = spacy.load(model_name)
            except OSError as e:
                raise OSError("Failed to get spaCy model. Download it manually using "
                              "`python -m spacy download %s`." % model_name) from e
        tokenizer[lang] = instance.tokenizer
        instance.tokenizer = lambda words: spacy.tokens.Doc(instance.vocab, words=words)
    return instance


nlp = {}
tokenizer = {}


def get_tokenizer(tokenized=False, lang="en"):
    instance = get_nlp(lang)
    return instance.tokenizer if tokenized else tokenizer[lang]


def get_word_vectors(dim=None, size=None, filename=None, lang="en"):
    vocab = get_nlp(lang).vocab
    if filename is not None:
        print("Loading word vectors from '%s'..." % filename)
        it = read_word_vectors(dim, size, filename)
        nr_row, nr_dim = next(it)
        i = 0
        if nr_row is None:
            vocab.reset_vectors(width=nr_dim)
        else:
            vocab.reset_vectors(shape=(nr_row, nr_dim))
        for word, vector in tqdm(it, total=nr_row, unit=" vectors", leave=False, mininterval=3):
            vocab_word = vocab[word]
            if not vocab_word.has_vector:
                vocab.set_vector(vocab_word.orth, np.asarray(vector[:nr_dim], dtype="f"))
                if vocab_word.has_vector:
                    i += 1
                    if nr_row and i >= nr_row:
                        break
    # elif dim is not None:  # Disabled due to explosion/spaCy#1518
    #     nr_row, nr_dim = vocab.vectors.shape
    #     if dim < nr_dim:
    #         vocab.vectors.resize(shape=(int(size or nr_row), int(dim)))
    lexemes = sorted([l for l in vocab if l.has_vector], key=attrgetter("prob"), reverse=True)[:size]
    return {l.orth_: l.vector for l in lexemes}, vocab.vectors_length


def read_word_vectors(dim, size, filename):
    try:
        first_line = True
        nr_row = nr_dim = None
        with open(filename, encoding="utf-8") as f:
            for line in f:
                fields = line.split()
                if first_line:
                    first_line = False
                    try:
                        nr_row, nr_dim = map(int, fields)
                        is_header = True
                    except ValueError:
                        nr_dim = len(fields) - 1  # No header, just get vector length from first one
                        is_header = False
                    if dim and dim < nr_dim:
                        nr_dim = dim
                    yield size or nr_row, nr_dim
                    if is_header:
                        continue  # Read next line
                word, *vector = fields
                if len(vector) >= nr_dim:  # May not be equal if word is whitespace
                    yield word, np.asarray(vector[-nr_dim:], dtype="f")
    except OSError as e:
        raise IOError("Failed loading word vectors from '%s'" % filename) from e


TAG_KEY = "tag"  # fine-grained POS tag
POS_KEY = "pos"  # coarse-grained POS tag
NER_KEY = "ner"  # named entity type
IOB_KEY = "iob"  # integer named entity IOB tag (0: unknown, 1: I, 2: O, 3: B)
DEP_KEY = "dep"  # dependency relation to syntactic head
HEAD_KEY = "head"  # integer position of syntactic head within paragraph (para_pos)
LEMMA_KEY = "lemma"
ANNOTATION_KEYS = (TAG_KEY, POS_KEY, NER_KEY, IOB_KEY, DEP_KEY, HEAD_KEY, LEMMA_KEY)


def annotate(passage, verbose=False, replace=False, lang="en"):
    """
    Run spaCy pipeline on the given passage
    :param passage: Passage object, whose layer 0 nodes will be added entries in the `extra' dict
    :param verbose: whether to print annotated text
    :param replace: even if a given passage is already annotated, replace with new annotation
    :param lang: optional two-letter language code
    """
    list(annotate_all([passage], verbose=verbose, replace=replace, lang=lang))


def annotate_all(passages, verbose=False, replace=False, lang="en"):
    """
    Run spaCy pipeline on the given passages
    :param passages: iterable of Passage objects, whose layer 0 nodes will be added entries in the `extra' dict
    :param verbose: whether to print annotated text
    :param replace: even if a given passage is already annotated, replace with new annotation
    :param lang: optional two-letter language code, will be overridden if passage has "lang" attrib
    :return generator of annotated passages, which are actually modified in-place (same objects as input)
    """
    for passage_lang, passages_by_lang in groupby(passages, get_lang):
        to_annotate = (([t.text for t in paragraph] if replace or not is_annotated(paragraph) else [],
                        (paragraph, passage))
                       for passage in passages_by_lang
                       for paragraph in break2paragraphs(passage, return_terminals=True))
        annotated = get_nlp(lang=passage_lang or lang).pipe(to_annotate, as_tuples=True)
        yield from (passage for passage, _ in groupby(apply_annotations(annotated, verbose)))


def get_lang(passage):
    return passage.attrib.get("lang")


def is_annotated(paragraph):
    return all(key in terminal.extra for terminal in paragraph for key in ANNOTATION_KEYS)


def apply_annotations(annotated, verbose):
    for doc, (paragraph, passage) in annotated:
        if doc:
            for lex, terminal in zip(doc, paragraph):
                terminal.extra[TAG_KEY] = lex.tag_
                terminal.extra[POS_KEY] = lex.pos_
                terminal.extra[NER_KEY] = lex.ent_type_
                terminal.extra[IOB_KEY] = str(lex.ent_iob)
                terminal.extra[DEP_KEY] = lex.dep_
                terminal.extra[HEAD_KEY] = str(lex.head.i + 1)
                terminal.extra[LEMMA_KEY] = lex.lemma_
        if verbose:
            extra = [["text"] + list(ANNOTATION_KEYS)] + \
                    [[t.text] + [t.extra[k] for k in ANNOTATION_KEYS] for t in paragraph]
            width = [max(len(f) for f in t) for t in extra]
            for i in range(1 + len(ANNOTATION_KEYS)):
                print(" ".join("%-*s" % (w, f[i]) for f, w in zip(extra, width)))
            print()
        yield passage


SENTENCE_END_MARKS = ('.', '?', '!')


def break2sentences(passage, lang="en"):
    """
    Breaks paragraphs into sentences according to the annotation.

    A sentence is a list of terminals which ends with a mark from
    SENTENCE_END_MARKS, and is also the end of a paragraph or parallel scene.
    :param passage: the Passage object to operate on
    :param lang: optional two-letter language code
    :return a list of positions in the Passage, each denotes a closing Terminal of a sentence.
    """
    l1 = passage.layer(layer1.LAYER_ID)
    terminals = extract_terminals(passage)
    if any(n.outgoing for n in l1.all):  # Passage is labeled
        ps_ends = [ps.end_position for ps in l1.top_scenes]
        ps_starts = [ps.start_position for ps in l1.top_scenes]
        marks = [t.position for t in terminals if t.text in SENTENCE_END_MARKS]
        # Annotations doesn't always include the ending period (or other mark)
        # with the parallel scene it closes. Hence, if the terminal before the
        # mark closed the parallel scene, and this mark doesn't open a scene
        # in any way (hence it probably just "hangs" there), it's a sentence end
        marks = [x for x in marks if x in ps_ends or ((x - 1) in ps_ends and x not in ps_starts)]
    else:  # Not labeled, split using spaCy
        annotated = get_nlp(lang=lang)([t.text for t in terminals])
        marks = [span.end for span in annotated.sents]
    marks = sorted(set(marks + break2paragraphs(passage)))
    # Avoid punctuation-only sentences
    if len(marks) > 1:
        marks = [x for x, y in zip(marks[:-1], marks[1:]) if not all(layer0.is_punct(t) for t in terminals[x:y])] + \
                [marks[-1]]
    return marks


def extract_terminals(p):
    """returns an iterator of the terminals of the passage p"""
    return p.layer(layer0.LAYER_ID).all


def break2paragraphs(passage, return_terminals=False):
    """
    Breaks into paragraphs according to the annotation.

    Uses the `paragraph' attribute of layer 0 to find paragraphs.
    :param passage: the Passage object to operate on
    :param return_terminals: whether to return actual Terminal objects of all terminals rather than just end positions
    :return a list of positions in the Passage, each denotes a closing Terminal of a paragraph.
    """
    terminals = list(extract_terminals(passage))
    return [list(p) for _, p in groupby(terminals, key=attrgetter("paragraph"))] if return_terminals else \
        [t.position - 1 for t in terminals if t.position > 1 and t.para_pos == 1] + [terminals[-1].position]


def indent_xml(xml_as_string):
    """
    Indents a string of XML-like objects.

    This works only for units with no text or tail members, and only for
    strings whose leaves are written as <tag /> and not <tag></tag>.
    :param xml_as_string: XML string to indent
    :return indented XML string
    """
    tabs = 0
    lines = str(xml_as_string).replace('><', '>\n<').splitlines()
    s = ''
    for line in lines:
        if line.startswith('</'):
            tabs -= 1
        s += ("  " * tabs) + line + '\n'
        if not (line.endswith('/>') or line.startswith('</')):
            tabs += 1
    return s



import sys

from ucca.ioutil import passage2file


def diff_passages(true_passage, pred_passage):
    """
    Debug method to print missing or mistaken attributes, nodes and edges
    """
    lines = list()
    if not true_passage._attrib.equals(pred_passage._attrib):
        lines.append("Passage attributes mismatch: %s, %s" %
                     (true_passage._attrib, pred_passage._attrib))
    try:
        for lid, l1 in true_passage._layers.items():
            l2 = true_passage.layer(lid)
            if not l1._attrib.equals(l2._attrib):
                lines.append("Layer %d attributes mismatch: %s, %s" %
                             (lid, l1._attrib, l2._attrib))
    except KeyError:  # no layer with same ID found
        lines.append("Missing layer: %s, %s" %
                     (true_passage._layers, pred_passage._layers))
    pred_ids = {node.extra["remarks"]: node
                for node in pred_passage.missing_nodes(true_passage)}
    true_ids = {node.ID: node
                for node in true_passage.missing_nodes(pred_passage)}
    for pred_id, pred_node in list(pred_ids.items()):
        true_node = true_ids.get(pred_id)
        if true_node:
            pred_ids.pop(pred_id)
            true_ids.pop(pred_id)
            pred_edges = {edge.tag + "->" + edge.child.ID: edge for edge in
                          pred_node.missing_edges(true_node)}
            true_edges = {edge.tag + "->" + edge.child.ID: edge for edge in
                          true_node.missing_edges(pred_node)}
            intersection = set(pred_edges).intersection(set(true_edges))
            pred_edges = {s: edge for s, edge in pred_edges.items() if s not in intersection}
            true_edges = {s: edge for s, edge in true_edges.items() if s not in intersection}

            node_lines = []
            if not pred_node._attrib.equals(true_node._attrib):
                node_lines.append("  Attributes mismatch: %s, %s" %
                                  (sorted(true_node._attrib.items()), sorted(pred_node._attrib.items())))
            if pred_edges:
                node_lines.append("  Mistake edges: %s" % ", ".join(pred_edges))
            if true_edges:
                node_lines.append("  Missing edges: %s" % ", ".join(true_edges))
            if node_lines:
                lines.append("For node " + pred_id + ":")
                lines.extend(node_lines)
    if pred_ids:
        lines.append("Mistake nodes: %s" % ", ".join(pred_ids))
    if true_ids:
        lines.append("Missing nodes: %s" % ", ".join(true_ids))
    if lines:
        outfile = "%s.xml" % true_passage.ID
        sys.stderr.write("Writing passage '%s'...\n" % outfile)
        passage2file(true_passage, outfile)
        outfile = "%s_pred.xml" % pred_passage.ID
        sys.stderr.write("Writing passage '%s'...\n" % outfile)
        passage2file(pred_passage, outfile)
    return "\n" + "\n".join(lines)



"""Input/output utility functions for UCCA scripts."""
import os
import sys
import time
from collections import defaultdict
from xml.etree.ElementTree import ParseError

from ucca.convert import file2passage, passage2file, from_text, to_text, split2segments
from ucca.core import Passage


class LazyLoadedPassages(object):
    """
    Iterable interface to Passage objects that loads files on-the-go and can be iterated more than once
    """
    def __init__(self, files, sentences=False, paragraphs=False, converters=None):
        self.files = files
        self.sentences = sentences
        self.paragraphs = paragraphs
        self.split = self.sentences or self.paragraphs
        self.converters = defaultdict(lambda: from_text) if converters is None else converters
        self._files_iter = None
        self._split_iter = None
        self._file_handle = None
        self._next_index = None

    def __iter__(self):
        self._next_index = 0
        self._files_iter = iter(self.files)
        self._split_iter = None
        self._file_handle = None
        return self

    def __next__(self):
        passage = self._next_passage()
        self._next_index += 1
        return passage

    def _next_passage(self):
        passage = None
        if self._split_iter is None:
            try:
                file = next(self._files_iter)
            except StopIteration:  # Finished iteration
                raise
            if isinstance(file, Passage):  # Not really a file, but a Passage
                passage = file
            else:  # A file
                attempts = 3
                while not os.path.exists(file):
                    if attempts == 0:
                        print("File not found: %s" % file, file=sys.stderr)
                        return next(self)
                    print("Failed reading %s, trying %d more times..." % (file, attempts), file=sys.stderr)
                    time.sleep(5)
                    attempts -= 1
                try:
                    passage = file2passage(file)  # XML or binary format
                except (IOError, ParseError):  # Failed to read as passage file
                    base, ext = os.path.splitext(os.path.basename(file))
                    converter = self.converters[ext.lstrip(".")]
                    self._file_handle = open(file, encoding="utf-8")
                    self._split_iter = iter(converter(self._file_handle, passage_id=base))
            if self.split:
                if self._split_iter is None:
                    self._split_iter = (passage,)
                self._split_iter = iter(s for p in self._split_iter for s in
                                        split2segments(p, is_sentences=self.sentences))
        if self._split_iter is not None:  # Either set before or initialized now
            try:
                # noinspection PyTypeChecker
                passage = next(self._split_iter)
            except StopIteration:  # Finished this converter
                self._split_iter = None
                if self._file_handle is not None:
                    self._file_handle.close()
                    self._file_handle = None
                return next(self)
        return passage

    # The following three methods are implemented to support shuffle;
    # note files are shuffled but there is no shuffling within files, as it would not be efficient.
    # Note also the inconsistency because these access the files while __iter__ accesses individual passages.
    def __len__(self):
        return len(self.files)

    def __getitem__(self, i):
        return self.files[i]

    def __setitem__(self, i, value):
        self.files[i] = value

    def __bool__(self):
        return bool(self.files)


def read_files_and_dirs(files_and_dirs, sentences=False, paragraphs=False, converters=None):
    """
    :param files_and_dirs: iterable of files and/or directories to look in
    :param sentences: whether to split to sentences
    :param paragraphs: whether to split to paragraphs
    :param converters: dict of input format converters to use based on the file extension
    :return: list of (lazy-loaded) passages from all files given,
             plus any files directly under any directory given
    """
    # TODO fix to a single list comprehension to preserve order; currently directories are moved to the end
    files = [files_and_dirs] if isinstance(files_and_dirs, str) else list(files_and_dirs)
    files += [os.path.join(d, f) for d in files if os.path.isdir(d) for f in os.listdir(d)]
    files = [f for f in files if not os.path.isdir(f)]
    return LazyLoadedPassages(files, sentences, paragraphs, converters)


def write_passage(passage, output_format=None, binary=False, outdir=".", prefix="", converter=None, verbose=True):
    suffix = output_format if output_format and output_format != "ucca" else ("pickle" if binary else "xml")
    outfile = outdir + os.path.sep + prefix + passage.ID + "." + suffix
    if verbose:
        print("Writing passage '%s'..." % outfile)
    if output_format is None or output_format in ("ucca", "pickle", "xml"):
        passage2file(passage, outfile, binary=binary)
    else:
        output = "\n".join(line for line in (converter or to_text)(passage))
        with open(outfile, "w", encoding="utf-8") as f:
            f.write(output + "\n")
    return outfile



#!/usr/bin/env python3
import argparse
import sys

from ucca.convert import from_json, CONVERTERS, TO_FORMAT
from ucca.ioutil import write_passage
from uccaapp.api import ServerAccessor

desc = """Download task from UCCA-App and convert to a passage in standard format"""


class TaskDownloader(ServerAccessor):
    def download_tasks(self, task_ids, **kwargs):
        for task_id in task_ids:
            yield self.download_task(task_id, **kwargs)

    def download_task(self, task_id, write=True, out_format=None, binary=None, out_dir=None, prefix=None, **kwargs):
        del kwargs
        passage = from_json(self.get_user_task(task_id), all_categories=self.layer["categories"])
        if write:
            write_passage(passage, out_format, binary, out_dir, prefix, TO_FORMAT.get(out_format))
        return passage

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("task_ids", nargs="+", type=int, help="IDs of tasks to download and convert")
        TaskDownloader.add_write_arguments(argparser)
        ServerAccessor.add_arguments(argparser)

    @staticmethod
    def add_write_arguments(argparser):
        argparser.add_argument("-f", "--out-format", choices=CONVERTERS, help="output file format (default: UCCA)")
        argparser.add_argument("-o", "--out-dir", default=".", help="output directory")
        argparser.add_argument("-p", "--prefix", default="", help="output filename prefix")
        argparser.add_argument("-b", "--binary", action="store_true", help="write in binary format (.pickle)")


def main(**kwargs):
    list(TaskDownloader(**kwargs).download_tasks(**kwargs))


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    TaskDownloader.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



from argparse import ArgumentParser

from ucca import constructions
from ucca.ioutil import read_files_and_dirs

if __name__ == "__main__":
    argparser = ArgumentParser(description="Extract linguistic constructions from UCCA corpus.")
    argparser.add_argument("passages", nargs="+", help="the corpus, given as xml/pickle file names")
    constructions.add_argument(argparser, False)
    argparser.add_argument("-v", "--verbose", action="store_true", help="print tagged text for each passage")
    args = argparser.parse_args()
    for passage in read_files_and_dirs(args.passages):
        if args.verbose:
            print("%s:" % passage.ID)
        extracted = constructions.extract_edges(passage, constructions=args.constructions, verbose=args.verbose)
        if any(extracted.values()):
            if not args.verbose:
                print("%s:" % passage.ID)
            for construction, edges in extracted.items():
                if edges:
                    print("  %s:" % construction.description)
                    for edge in edges:
                        print("    %s [%s %s]" % (edge, edge.tag, edge.child))
            print()
